// Title:	BiDiPro DLL module
// Desc:	SPI EEPROM
// Ver:		0.0.1
// Found:	2010-03-18
// Last:	2012-09-11
// CPU:		386+
// Model:	Tiny
// Compil:	Borland C++ 3.1 (DOS)
// Author:	Oleg O. Chukaev aka 0leg


/****
25C BiDiPro module
Copyright (c) 2012, 㪠   <oleg.chukaev@mail.ru>

蠥 ୮ ࠭  ᯮ짮    室
, ⠪   筮 ଥ,  ﬨ  ,  ᮡ ᫥
᫮:

1.  ୮ ࠭ 室   ⠢ 㪠
    㢥  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩.
2.  ୮ ࠭ 筮   ࠭ 㪠
    ଠ  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩  㬥樨 /  㣨 ਠ, ⠢塞
    ࠭.

      /  
"  "  -  ,   
, ,    ,  
      .    ,
    ,      ,
         ,   
/   ,    ,  
,   , ,   
,      
(,     ,  , 
,    -    , 
      ),   
         .
****/


#include <stdlib.h>
#include "../dlldata.h"
#include "../errors.h"
#include "../bidi.h"


#ifdef LINUXXX
 extern DLLDATA *dlp;
#else
# define dlp ((DLLDATA *) NULL)		/* ⥫    */
#endif

#define bdport dlp->bdprt		/* ⥫  hardware ࠩ */

typedef struct CHDEFS {
	dword size;	// 𬪮
	byte addrmode;	//  樨
} CHDEFS;

CHDEFS chd[]={
//	size        addrmode
	{0x000001L, 0},		//Status Register
	{0x000080L, 1},		//010
	{0x000100L, 1},		//020
	{0x000200L, 1},		//040
	{0x000400L, 2},		//080
	{0x000800L, 2},		//160
	{0x001000L, 2},		//320
	{0x002000L, 2},		//640
	{0x004000L, 2},		//128
	{0x008000L, 2},		//256
	{0x010000L, 2},		//512
	{0x020000L, 3},		//1024
};

// ७   CFG:
static struct CFGKEY cfgkeys[] = {
	{'D', "wpage", 1, 1, 65536L},			//Page size
	{'D', "wtime", 20, 20, 100000L},		//Write time (us)
	{'D', "etime", 0, 0, 40000000L},		//Erase time (us)
	{'D', "id_alg", 0, 0, 1},			//Read ID alg.
	{'D', "er_alg", 0, 0, 1},			//Erase alg.
	{0, {0}, 0, 0},					//End of table
};

#define cfg_wpage	cfgkeys[0].data
#define cfg_wtime	cfgkeys[1].data
#define cfg_etime	cfgkeys[2].data
#define	cfg_id_alg	cfgkeys[3].data
#define	cfg_er_alg	cfgkeys[4].data

/* Output, 1B */
#define PIN_CE		0x01
#define PIN_WP		0x04
#define PIN_SI		0x08
#define PIN_SCK		0x10
#define PIN_HOLD	0x20

/* Input, 1A */
#define PIN_SO		0x01

#define SPI_WREN	0x06
#define SPI_WRDI	0x04
#define SPI_RDSR	0x05
#define SPI_WRSR	0x01
#define SPI_READ	0x03
#define SPI_WRITE	0x02
#define	SPI_PE		0x42
#define	SPI_SE		0xd8
#define	SPI_CE		0xc7
#define	SPI_RDID	0xab

#define WRSR_TIME	150000L		/* Delay for WRSR command */


static char *copyright = "25C 0.0.1  SPI EEPROM module  "
	"(c) Oleg O. Chukaev  ["__DATE__"]";
static word chtype = 0xffff;
static CHDEFS *pchd;
static byte pin_wp = 0;


/*
 | 몫祭 ⠭
 */
static void power_off (void) {
	bdport (PRST);
}


/*
 | 祭 ⠭
 */
static void power_on (void) {

	bdport (PFORCE | PWR1 | PAIN);		/* DD1:IOO */
	bdport (PFORCE | PWR2);			/* DD4:OOO */
	dlp->voltage (dlp->Vcc, 0, 0);
	bdport (PWB2 | 0x80);			/* Vcc pin 36 */
	bdport (PWA2 | BITTER);
	bdport (PWB1 | PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (10000);
}


/*
 | Get value of one pin
 |
 | Ret: 0 / 1
 */
static byte get_pin (byte pin) {
	return (bdport (PRA1) & pin) ? 1 : 0;
}


/*
 | Set all pins
 */
static void set_pins (byte pins) {
	bdport (PWB1 | pins);
}


static void pin_ce_1 (void) {
  	set_pins (PIN_HOLD | pin_wp | PIN_CE);
}


static void pin_ce_0 (void) {
  	set_pins (PIN_HOLD | pin_wp);
}


/*
 |  
 */
static byte spi_read_byte (void) {

	word i;
	byte d = 0;

	for (i = 0; i < 8; i++) {
		d <<= 1;
		d |= get_pin (PIN_SO);
		set_pins (pin_wp | PIN_HOLD | PIN_SCK);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);		/* SCK=0 */
	}
	return d;
}


/*
 | । 
 */
static void spi_write_byte (byte d) {

	word i;
	byte z;

	for (i = 0; i < 8; i++) {
		z = (d & 0x80) ? PIN_SI : 0;
		set_pins (pin_wp | PIN_HOLD | z);		/* data */
		set_pins (pin_wp | PIN_HOLD | PIN_SCK | z);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);			/* SCK=0 */
		d <<= 1;
	}
}


/*
 | । 8-, 16-  24-⭮ 
 */
static void spi_write_cmd_addr (byte cmd, dword addr) {

	if (pchd->addrmode == 1) {
		spi_write_byte (((addr << 3L) & 0x0800) | cmd);	/* cmd + A8 */
	} else if (pchd->addrmode == 2) {
		spi_write_byte (cmd);				/* cmd */
		spi_write_byte ((addr >> 8L) & 0xff);		/* A15..A8 */
	} else if (pchd->addrmode == 3) {
		spi_write_byte (cmd);				/* cmd */
		spi_write_byte ((addr >> 16L) & 0xff);		/* A23..A16 */
		spi_write_byte ((addr >> 8L) & 0xff);		/* A15..A8 */
	}
	spi_write_byte ((addr >> 0L) & 0xff);		/* A7..A0 */
}


/*
 | Send Write Enable command
 |
 | CE = 1 on exit
 */
static void send_wren (void) {

	pin_ce_0 ();
	spi_write_byte (SPI_WREN);
	pin_ce_1 ();
}


/*
 | Read Status Register
 |
 | CE = 1 on exit
 */
static byte read_status_reg (void) {

	byte d;

	pin_ce_0 ();			//!!!! s_w_b  CE  0...
	spi_write_byte (SPI_RDSR);
	d = spi_read_byte ();
	pin_ce_1 ();
	return d;
}


/*
 | Write Status Register
 |
 | CE = 1 on exit
 */
static void write_status_reg (byte d) {

	pin_ce_0 ();
	spi_write_byte (SPI_WREN);
	pin_ce_1 ();
	dlp->mcsdelay (1000);		//!!!!
	pin_ce_0 ();			//!!!!
	spi_write_byte (SPI_WRSR);
	spi_write_byte (d);
	pin_ce_1 ();
	dlp->mcsdelay (WRSR_TIME);
}


/* Unprotect blocks */
static void unprotect (void) {

	set_pins (PIN_HOLD | pin_wp);
	spi_write_byte (SPI_WREN);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (1000);
	set_pins (PIN_HOLD | pin_wp);
	spi_write_byte (SPI_WRSR);
	spi_write_byte (0x00);		/* clear BP0, BP1, ... */
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (50000L);
}


/*
 | Poll BUSY/WIP bit
 */
static int poll_busy_bit (dword timeout) {

	dword i;

	for (i = 0; i <= timeout / 20; i++) {
		dlp->mcsdelay (20);
		if ((read_status_reg () & 0x01) == 0x00) {
			return 0;
		}
	}
	return 1;
}


/*
 | Compare page
 */
static int compare_page (dword addr) {

	dword i;
	byte d;

	spi_write_cmd_addr (SPI_READ, addr);
	for (i = 0; i < cfg_wpage; i++) {
		d = spi_read_byte ();
		if (dlp->getbytebuf (addr + i) != d) {
			dlp->adr = addr + i;
			dlp->edata = d;
			return 1;
		}
	}
	return 0;
}


/*
 | Read, compare, blank check
 */
static int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	dword errors = 0;
	byte dat;

	/* Status Register */
	if (pchd->size == 1) {
		power_on ();
		dat = read_status_reg ();
		power_off ();
		dlp->setbytebuf (0, dat);
		return 0;
	}

	/* EEPROM */
	power_on ();
	pin_ce_0 ();
	spi_write_cmd_addr (SPI_READ, 0);
	for (adr = 0; adr < pchd->size; adr++) {
		dat = spi_read_byte ();
		switch (dlp->funcode) {
		case DFN_FREAD:
			dlp->setbytebuf (adr, dat);
			break;

		case DFN_FCOMPARE:
			if (dlp->getbytebuf (adr) != dat) {
				/* ⬥ ᮢ */
				dlp->setbmark (adr, 1);
				errors++;
			}
			break;

		case DFN_FBLANK:
			if (dat != 0xff) {
				dlp->adr = adr;
				power_off ();
				return ERC_BLANK;
			}
			break;
		}
		if (adr % istep == 0 && dlp->iprogress (adr, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return 0;
}


/*
 | Read ID
 */
static int read_id (void) {

	dword id1 = 0;

	power_on ();
	pin_ce_0 ();
	spi_write_cmd_addr (SPI_RDID, 0x0000);
	id1 = spi_read_byte ();
	power_off ();
	dlp->edata = id1;
	return 0;
}


/*
 | Erase
 */
static int erase (void) {

	dword istep;
	dword i;
	dword w0;
	byte saved_sr;

	pin_wp = PIN_WP;
	power_on ();
	saved_sr = read_status_reg ();
	unprotect ();
	send_wren ();
	pin_ce_0 ();
	spi_write_byte (SPI_CE);
	pin_ce_1 ();
	if (cfg_etime == 0L) {
		/* poll BUSY bit */
		dlp->mcsdelay (10000);
		dlp->iprogress (1600, 0);		/* init, 16 s */
		istep = 1600 / 64;
		i = 0;
		do {
			dlp->mcsdelay (10000);
			w0 = read_status_reg ();
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			i++;
		} while ((w0 & 0x01) == 0x01);
		pin_ce_1 ();
	} else if (cfg_etime <= 300000L) {
		/* <= 0.3 s */
		dlp->mcsdelay (cfg_etime);		/* delay */
	} else {
		/* etime */
		dlp->iprogress (cfg_etime / 1000, 0);	/* init */
		istep = cfg_etime / 64000L;
		for (i = 0; i < cfg_etime / 1000; i++) {
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			dlp->mcsdelay (1000);		/* delay */
		}
	}
	dlp->iprogress (0, 1);				/* close */
	write_status_reg (saved_sr);
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Write
 */
static int write_all (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size / 64L;
	byte saved_sr;

	/* Status Register */
	if (pchd->size == 1) {
		pin_wp = PIN_WP;
		power_on ();
		write_status_reg (dlp->getbytebuf (0));
		power_off ();
		pin_wp = 0;
		return 0;
	}

	if (dlp->checkwr) {
		dlp->iprogress ((dword)"Writing, verifying chip", 3);
	} else {
		dlp->iprogress ((dword)"Writing chip", 3);
	}
	dlp->iprogress (pchd->size, 0);

	/* EEPROM */
	pin_wp = PIN_WP;
	power_on ();
	saved_sr = read_status_reg ();
	unprotect ();
	send_wren ();
	pin_ce_0 ();

	for (adr = 0; adr < pchd->size; adr += cfg_wpage) {
		spi_write_cmd_addr (SPI_WRITE, adr);
		for (i = 0; i < cfg_wpage; i++) {
			spi_write_byte (dlp->getbytebuf (adr+i));
		}
		pin_ce_1 ();
//		dlp->mcsdelay (cfg_wtime * cfg_wpage);
		if (poll_busy_bit (cfg_wtime * cfg_wpage)) {
			write_status_reg (saved_sr);
			power_off ();
			pin_wp = 0;
			dlp->adr = adr;
			return ERC_WRITE;
		}
		if (dlp->checkwr) {
			if (compare_page (adr)) {
				write_status_reg (saved_sr);
				power_off ();
				pin_wp = 0;
				return ERC_WRITE;
			}
			pin_ce_1 ();
		}
		send_wren ();
		if ((adr + cfg_wpage) % istep == 0
				&& dlp->iprogress (adr + cfg_wpage, 2)) {
			break;
		}
	}

	write_status_reg (saved_sr);
	power_off ();
	pin_wp = 0;
	return 0;
}


/* ⠭ ⨯ ROM */
static word settype (void) {

	word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize) {
			if (chd[i].size == 1) {
				/* Status Register */
				dlp->funmask[DFN_FCOMPARE] = 0;
				dlp->funmask[DFN_FBLANK] = 0;
			} else {
				/* EEPROM */
				dlp->funmask[DFN_FCOMPARE] = 1;
				dlp->funmask[DFN_FBLANK] = 1;
			}
			dlp->funmask[DFN_READID] = (cfg_id_alg != 0);
			dlp->funmask[DFN_ERASE] = (cfg_er_alg != 0);
			dlp->funmask[DFN_FREAD] = 1;
			dlp->funmask[DFN_FWRITE] = 1;
			return (chtype = i);
		}
	}
	return (chtype = 0xffff);
}


// main code
int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⠫ -樨 ⨢  settype(). */
		dlp->funmask[DFN_INIT] = 1;
		dlp->funmask[DFN_FREAD] = 0;
		dlp->funmask[DFN_FWRITE] = 0;
		dlp->funmask[DFN_TEST] = 0;
		dlp->funmask[DFN_LOCK] = 0;
		dlp->defstate = 0xff;
		dlp->copyright = (byte *)copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		return 0;

	case DFN_INIT:
		if (settype () == 0xffff)
			return ERC_CHIPUNSP;
		pchd = &chd[chtype];
		dlp->dippins = 8;
		return 0;

	case DFN_FREAD:
		return read_all ();

	case DFN_FWRITE:
		return write_all ();

	case DFN_FCOMPARE:
		return read_all ();

	case DFN_FBLANK:
		return read_all ();

	case DFN_READID:
		return read_id ();

	case DFN_ERASE:
		return erase ();

	default:
		return ERC_FUNUNSP;
	}
}

/* E0F */

